require "draw_primitive.rb"


# --------------------------------------------------------
#  line mixing
# --------------------------------------------------------
module LineCrossCheck
  def crossbase(line1, line2)
    result = Hash.new
    l1p1 = line1.get_point(0)
    l1p2 = line1.get_point(1)
    l2p1 = line2.get_point(0)
    l2p2 = line2.get_point(1)
    det = (l2p2.x - l2p1.x)*(l1p2.y - l1p1.y) - (l1p2.x - l1p1.x)*(l2p2.y - l2p1.y)
    if det == 0 then
      return nil
    end
    dx = l2p1.x - l1p1.x
    dy = l2p1.y - l1p1.y
    t1 = ((l2p2.x - l2p1.x)*dy - (l2p2.y - l2p1.y)*dx) / det
    t2 = ((l1p2.x - l1p1.x)*dy - (l1p2.y - l1p1.y)*dx) / det
    t1 = 0 if t1 < 0.000000001 && t1 > 0
    t2 = 0 if t2 < 0.000000001 && t1 > 0
    t1 = 1.0 if t1 > 0.999999999 && t1 < 1.0
    t2 = 1.0 if t2 > 0.999999999 && t1 < 1.0
    result['t1'] = t1
    result['t2'] = t2
    return result
  end

  def crosspoint(line1, line2)
    t = crossbase(line1, line2)
    l1p1 = line1.get_point(0)
    l1p2 = line1.get_point(1)
    l2p1 = line2.get_point(0)
    l2p2 = line2.get_point(1)
    crosspoint = PPoint.new
    crosspoint.x = l1p1.x + (l1p2.x - l1p1.x) * t['t1']
    crosspoint.y = l1p1.y + (l1p2.y - l1p1.y) * t['t1']
    crosspoint.z = 0
    return crosspoint
  end

  def cross?(line1, line2)
    t = crossbase(line1, line2)
    if t == nil then
      return false
    end
    return ((t['t1'] > 0) && (t['t1'] < 1.0) && (t['t2'] > 0) && (t['t2'] < 1.0))
  end

  def cross_with_top?(line1, line2)
    t = crossbase(line1, line2)
    if t == nil then
      return false
    end
    return ((t['t1'] >= 0) && (t['t1'] <= 1.0) && (t['t2'] >= 0) && (t['t2'] <= 1.0))
  end

  def online?(line, point)
    result = false
    l1p1 = line.get_point(0)
    l1p2 = line.get_point(1)
    result=true if l1p1.equal?(point)
    result=true if l1p2.equal?(point)
    v1 = PPoint.new(l1p2.x-l1p1.x, l1p2.y-l1p1.y)
    v2 = PPoint.new(point.x-l1p1.x, point.y-l1p1.y)
    inner = v1.inner_product(v2)
    if (inner > 0.99999999) && (v1.length >= v2.length) then
      result = true
    end
    return result
  end

end



# --------------------------------------------------------
#  pattern stored class
# --------------------------------------------------------

class PatternStorage

  def initialize
    @object = []
    @outline = []
  end

  attr_reader :object
  attr_reader :outline

  def object_size?
    @object.size
  end

  def clear
    @object.clear
    @outline.clear
  end

  def add( triangle )
    triangle.normalize()
    if @object.size == 0 then
      @object << triangle
    else
      marge( triangle )
    end
  end

  def drip_outline()
    @outline = []
    make_outline()
  end

  # ------------------------------
  # --- object marge work
  def marge( addtri )
    @object << addtri
  end

  # ------------------------------
  # --- extruct object outline
  class PLineCount < PLine
    def initialize
      super
      @count = 0
    end

    attr_accessor :count
  end

  def make_outline()
    @outline = []
    linelist = []
    @object.each do |tri|
      for i in 0...3
        p1 = tri.get_point(i)
        p2 = tri.get_point((i+1)%3)
        line = PLineCount.new
        line.set_point(0, p1)
        line.set_point(1, p2)
        check = false
        linelist.each do |ol|
          if ol.same?(line) then
            ol.count += 1
            check = true
            break
          end
        end
        linelist << line if check==false
      end
    end
    # - pickup outline
    linelist.each do |ol|
      if ol.count == 0 then
        line = PLine.new
        line.set_point(0, ol.get_point(0))
        line.set_point(1, ol.get_point(1))
        @outline << line
      end
    end
  end

end



# --------------------------------------------------------
#  draw pattern primitive
# --------------------------------------------------------

class PatternObject
  include LineCrossCheck

  def initialize
    @pattern = PatternStorage.new
    @mil_path = PatternStorage.new
    @mil_outline = []
    @sequence = []
    @mil_size = 0.0
    @pad_size = 0.5
    @flash_pattern = 0
    @circle_resolution = 4
    @bounding_box = { 'x1'=>0,'y1'=>0,'x2'=>0,'y2'=>0 }
  end

  attr_accessor :circle_resolution
  attr_accessor :bounding_box

  def clear_pattern
    @pattern.clear
    @mil_path.clear
  end

  def clear_sequence
    @sequence.clear
  end

  def set_mil_size(size)
    @mil_size = size
  end

  def set_pad_size(size)
    @pad_size = size
  end

  def set_pad_round(type)
    if type == true then
      # round pattern
      @flash_pattern = 1
    else
      # square pattern
      @flash_pattern = 0
    end
  end

  def get_pattern_object
    @pattern.object
  end

  def get_mil_path
    @mil_outline
  end

  def set_mil_path(lines)
    @mil_outline = lines
  end

  def drip_mil_path
    @mil_path.drip_outline()
    @mil_outline = @mil_path.outline
    calc_bounding_box
  end

  # -----------------------------
  # --- do plot

  def flash(x, y)
    if @flash_pattern == 1 then
      draw_circle_plot(x, y, x, y, @pad_size/2, @pattern)
      draw_circle_plot(x, y, x, y, (@pad_size+@mil_size)/2, @mil_path)
    else
      draw_cube_plot(x, y, x, y, @pad_size/2, @pattern)
      draw_cube_plot(x, y, x, y, (@pad_size+@mil_size)/2, @mil_path)
    end
  end

  def move(x1, y1, x2, y2)
    if @flash_pattern == 1 then
      draw_circle_plot(x1, y1, x2, y2, @pad_size/2, @pattern)
      draw_circle_plot(x1, y1, x2, y2, (@pad_size+@mil_size)/2, @mil_path)
    else
      draw_cube_plot(x1, y1, x2, y2, @pad_size/2, @pattern)
      draw_cube_plot(x1, y1, x2, y2, (@pad_size+@mil_size)/2, @mil_path)
    end
  end

  # -----------------------------
  # --- bounding box

  def calc_bounding_box
    if @mil_outline.size > 0 then
      line = @mil_outline[0]
      @bounding_box['x1'] = line[0].x
      @bounding_box['y1'] = line[0].y
      @bounding_box['x2'] = line[1].x
      @bounding_box['y2'] = line[1].y
      if @bounding_box['x1'] > @bounding_box['x2'] then
        t = @bounding_box['x1']
        @bounding_box['x1'] = @bounding_box['x2']
        @bounding_box['x2'] = t
      end
      if @bounding_box['y1'] > @bounding_box['y2'] then
        t = @bounding_box['y1']
        @bounding_box['y1'] = @bounding_box['y2']
        @bounding_box['y2'] = t
      end
      for i in 1...@mil_outline.size
        line = @mil_outline[i]
        for j in 0..1
          if @bounding_box['x1'] > line[j].x then
            @bounding_box['x1'] = line[j].x
          end
          if @bounding_box['y1'] > line[j].y then
            @bounding_box['y1'] = line[j].y
          end
          if @bounding_box['x2'] < line[j].x then
            @bounding_box['x2'] = line[j].x
          end
          if @bounding_box['y2'] < line[j].y then
            @bounding_box['y2'] = line[j].y
          end
        end
      end
    else
      @bounding_box = { 'x1'=>0,'y1'=>0,'x2'=>0,'y2'=>0 }
    end
  end

  def collision_bounding_box?(box)
    # - include
    if ((@bounding_box['x1'] <= box['x1']) && (box['x1'] <= @bounding_box['x2'])) ||
       ((@bounding_box['x1'] <= box['x2']) && (box['x2'] <= @bounding_box['x2'])) then
      if ((@bounding_box['y1'] <= box['y1']) && (box['y1'] <= @bounding_box['y2'])) ||
         ((@bounding_box['y1'] <= box['y2']) && (box['y2'] <= @bounding_box['y2'])) then
        return true
      end
    end
    if ((box['x1'] <= @bounding_box['x1']) && (@bounding_box['x1'] <= box['x2'])) ||
       ((box['x1'] <= @bounding_box['x2']) && (@bounding_box['x2'] <= box['x2'])) then
      if ((box['y1'] <= @bounding_box['y1']) && (@bounding_box['y1'] <= box['y2'])) ||
         ((box['y1'] <= @bounding_box['y2']) && (@bounding_box['y2'] <= box['y2'])) then
        return true
      end
    end
    # - closs
    myline = []
    myline[0] = PLine.new
    myline[1] = PLine.new
    myline[2] = PLine.new
    myline[3] = PLine.new
    myline[0].set_point(0, PPoint.new(@bounding_box['x1'], @bounding_box['y1']))
    myline[0].set_point(1, PPoint.new(@bounding_box['x2'], @bounding_box['y1']))
    myline[1].set_point(0, PPoint.new(@bounding_box['x2'], @bounding_box['y1']))
    myline[1].set_point(1, PPoint.new(@bounding_box['x2'], @bounding_box['y2']))
    myline[2].set_point(0, PPoint.new(@bounding_box['x2'], @bounding_box['y2']))
    myline[2].set_point(1, PPoint.new(@bounding_box['x1'], @bounding_box['y2']))
    myline[3].set_point(0, PPoint.new(@bounding_box['x1'], @bounding_box['y2']))
    myline[3].set_point(1, PPoint.new(@bounding_box['x1'], @bounding_box['y1']))
    oline = []
    oline[0] = PLine.new
    oline[1] = PLine.new
    oline[2] = PLine.new
    oline[3] = PLine.new
    oline[0].set_point(0, PPoint.new(box['x1'], box['y1']))
    oline[0].set_point(1, PPoint.new(box['x2'], box['y1']))
    oline[1].set_point(0, PPoint.new(box['x2'], box['y1']))
    oline[1].set_point(1, PPoint.new(box['x2'], box['y2']))
    oline[2].set_point(0, PPoint.new(box['x2'], box['y2']))
    oline[2].set_point(1, PPoint.new(box['x1'], box['y2']))
    oline[3].set_point(0, PPoint.new(box['x1'], box['y2']))
    oline[3].set_point(1, PPoint.new(box['x1'], box['y1']))
    myline.each do |line1|
      oline.each do |line2|
        if cross?(line1, line2) then
          return true
        end
      end
    end
    #
    return false
  end


  # -----------------------------
  # --- plot line

  def get_side(x1, y1, x2, y2, size)
    # --- line angle
    if (x1 != x2) || (y1 != y2) then
      line_vector = PPoint.new(x2-x1, y2-y1, 0)
    else
      line_vector = PPoint.new(1.0, 0, 0)
    end
    line_vector.normalize
    line_vector.zoom(size)
    line_vector.rotate_z(-90.0)
    return line_vector
  end

  def draw_line(x1, y1, x2, y2, size, storage)
    width = get_side(x1, y1, x2, y2, size)
    # --- box1
    p1 = PPoint.new(x1, y1, 0)
    p2 = PPoint.new(x2, y2, 0)
    p3 = PPoint.new(x2, y2, 0)
    p4 = PPoint.new(x1, y1, 0)
    p3.add(width)
    p4.add(width)
    tri1 = PTriangle.new
    tri1.set_point(0, p1)
    tri1.set_point(1, p2)
    tri1.set_point(2, p3)
    tri1.normalize
    tri2 = PTriangle.new
    tri2.set_point(0, p3)
    tri2.set_point(1, p4)
    tri2.set_point(2, p1)
    tri2.normalize
    storage.add(tri1)
    storage.add(tri2)
    # --- box2
    width2 = PPoint.new(width.x, width.y, width.z)
    width2.rotate_z(180.0)
    p1 = PPoint.new(x1, y1, 0)
    p2 = PPoint.new(x2, y2, 0)
    p3 = PPoint.new(x2, y2, 0)
    p4 = PPoint.new(x1, y1, 0)
    p3.add(width2)
    p4.add(width2)
    tri1 = PTriangle.new
    tri1.set_point(0, p1)
    tri1.set_point(1, p2)
    tri1.set_point(2, p4)
    tri1.normalize
    tri2 = PTriangle.new
    tri2.set_point(0, p3)
    tri2.set_point(1, p4)
    tri2.set_point(2, p2)
    tri2.normalize
    storage.add(tri1)
    storage.add(tri2)
    # --- box2
  end

  def draw_cube_plot(x1, y1, x2, y2, size, storage)
    width = get_side(x1, y1, x2, y2, size)
    width.rotate_z(-90.0)
    p1 = PPoint.new(x1, y1, 0)
    p1.add(width)
    draw_line(x1, y1, p1.x, p1.y, size, storage)
    width.rotate_z(180.0)
    p2 = PPoint.new(x2, y2, 0)
    p2.add(width)
    draw_line(x2, y2, p2.x, p2.y, size, storage)
    #
    if (x1 != x2) || (y1 != y2) then
      draw_line(x1, y1, x2, y2, size, storage)
    end
  end

  def draw_circle_plot(x1, y1, x2, y2, size, storage)
    div = @circle_resolution
    width1 = get_side(x1, y1, x2, y2, size)
    width2 = PPoint.new(width1.x, width1.y, width1.z)
    width2.rotate_z(180.0)
    dangle = -(180.0 / div)
    #
    mp = PPoint.new(width1.x, width1.y, width1.z)
    pp = PPoint.new(width1.x, width1.y, width1.z)
    zp = PPoint.new(x1, y1, 0)
    for i in 0...(div-1)
      mp.rotate_z(dangle)
      p1 = PPoint.new(mp.x, mp.y, mp.z)
      p2 = PPoint.new(pp.x, pp.y, pp.z)
      p1.add(zp)
      p2.add(zp)
      tri = PTriangle.new
      tri.set_point(0, zp)
      tri.set_point(1, p1)
      tri.set_point(2, p2)
      tri.normalize
      storage.add(tri)
      pp = PPoint.new(mp.x, mp.y, mp.z)
    end
    p1 = PPoint.new(mp.x, mp.y, mp.z)
    p2 = PPoint.new(width2.x, width2.y, width2.z)
    p1.add(zp)
    p2.add(zp)
    tri = PTriangle.new
    tri.set_point(0, zp)
    tri.set_point(1, p1)
    tri.set_point(2, p2)
    tri.normalize
    storage.add(tri)
    #
    mp = PPoint.new(width2.x, width2.y, width2.z)
    pp = PPoint.new(width2.x, width2.y, width2.z)
    zp = PPoint.new(x2, y2, 0)
    for i in 0...(div-1)
      mp.rotate_z(dangle)
      p1 = PPoint.new(mp.x, mp.y, mp.z)
      p2 = PPoint.new(pp.x, pp.y, pp.z)
      p1.add(zp)
      p2.add(zp)
      tri = PTriangle.new
      tri.set_point(0, zp)
      tri.set_point(1, p1)
      tri.set_point(2, p2)
      tri.normalize
      storage.add(tri)
      pp = PPoint.new(mp.x, mp.y, mp.z)
    end
    p1 = PPoint.new(mp.x, mp.y, mp.z)
    p2 = PPoint.new(width1.x, width1.y, width1.z)
    p1.add(zp)
    p2.add(zp)
    tri = PTriangle.new
    tri.set_point(0, zp)
    tri.set_point(1, p1)
    tri.set_point(2, p2)
    tri.normalize
    storage.add(tri)
    #
    if (x1 != x2) || (y1 != y2) then
      draw_line(x1, y1, x2, y2, size, storage)
    end
  end

  # -----------------------------
  # --- marge pattern outline

  def marge_outline(item, master)
    # --- keep same line
    protect_line = []
    item.each do |s_line|
      @mil_outline.each do |m_line|
        if m_line.same?(s_line) then
          protect_line << s_line
          break
        end
      end
    end
    # --- predrop include line
    dropitem = []
    item.each do |line|
      if point_include?(line) then
        dropitem << line
      end
    end
    item -= dropitem
    # --- clip cross line
    chunk = []
    dropitem = []
    item.each do |line1|
      s = split_line(line1)
      if s.size > 1 then
        dropitem << line1
        chunk += s
      end
    end
    item -= dropitem
    item += chunk
    # --- drop include line
    dropitem = []
    item.each do |line|
      if point_include?(line) then
        dropitem << line
      end
    end
    item -= dropitem
    # --- protect for master outline
    if master == true then
      item += protect_line
    else
      item -= protect_line
    end
    #
    return item
  end

  def cross_line?(item)
    result = false
    item.each do |line1|
      @mil_path.object.each do |tri|
        line2 = PLine.new
        line2.set_point(0, tri.get_point(0))
        line2.set_point(1, tri.get_point(1))
        if cross?(line1, line2) then
          result = true
        end
        line2.set_point(0, tri.get_point(1))
        line2.set_point(1, tri.get_point(2))
        if cross?(line1, line2) then
          result = true
        end
        line2.set_point(0, tri.get_point(2))
        line2.set_point(1, tri.get_point(0))
        if cross?(line1, line2) then
          result = true
        end
        if (tri[0] == line1[0]) || (tri[1] == line1[0]) || (tri[2] == line1[0]) then
          result = true
        end
        if (tri[0] == line1[1]) || (tri[1] == line1[1]) || (tri[2] == line1[1]) then
          result = true
        end
      end
      break if result
    end
    return result
  end

  def item_include?(item)
    result = false
    item.each do |line|
      if point_include?(line) then
        result = true
        break
      end
    end
    return result
  end

  def point_include?(line)
    result1 = false
    result2 = false
    @mil_path.object.each do |tri|
      if tri.include?(line.get_point(0)) then
        result1 = true
      end
      if tri.include?(line.get_point(1)) then
        result2 = true
      end
      l = PLine.new
      l.set_point(0, tri.get_point(0))
      l.set_point(1, tri.get_point(1))
      if online?(l, line.get_point(0)) then
        result1 = true
      end
      if online?(l, line.get_point(1)) then
        result2 = true
      end
      l.set_point(0, tri.get_point(1))
      l.set_point(1, tri.get_point(2))
      if online?(l, line.get_point(0)) then
        result1 = true
      end
      if online?(l, line.get_point(1)) then
        result2 = true
      end
      l.set_point(0, tri.get_point(2))
      l.set_point(1, tri.get_point(0))
      if online?(l, line.get_point(0)) then
        result1 = true
      end
      if online?(l, line.get_point(1)) then
        result2 = true
      end
      if (tri[0]==line[0])||(tri[1]==line[0])||(tri[2]==line[0]) then
        result1 = true
      end
      if (tri[0]==line[1])||(tri[1]==line[1])||(tri[2]==line[1]) then
        result2 = true
      end
    end
    return (result1==true && result2==true)
  end

  def cutting(line1, line2)
    clip = []
    xp = crosspoint(line1, line2)
    nl1 = PLine.new
    nl1.set_point(0, line1.get_point(0))
    nl1.set_point(1, xp)
    nl2 = PLine.new
    nl2.set_point(0, xp)
    nl2.set_point(1, line1.get_point(1))
    clip << nl1
    clip << nl2
    return clip
  end

  def split_line(cline)
    divline = []
    divline << cline

    if cline.length < 0.001 then
      return divline
    end

    @mil_path.object.each do |tri|
      line1 = PLine.new
      line1.set_point(0, tri.get_point(0))
      line1.set_point(1, tri.get_point(1))
      line2 = PLine.new
      line2.set_point(0, tri.get_point(1))
      line2.set_point(1, tri.get_point(2))
      line3 = PLine.new
      line3.set_point(0, tri.get_point(2))
      line3.set_point(1, tri.get_point(0))
      #
      hit1 = cross?(cline, line1)
      hit2 = cross?(cline, line2)
      hit3 = cross?(cline, line3)
      if ((hit1 == false) &&
          (hit2 == false) &&
          (hit3 == false)) then
        next
      end
      #
      if (hit1 == true) then
        newline = []
        divline.each do |line|
          if cross?(line, line1) then
            splitline = cutting(line, line1)
            newline << splitline[0] if splitline[0].length > 0.0 
            newline << splitline[1] if splitline[1].length > 0.0 
          else 
            newline << line
          end
        end
        divline = newline
      end
      if (hit2 == true) then
        newline = []
        divline.each do |line|
          if cross?(line, line2) then
            splitline = cutting(line, line2)
            newline << splitline[0] if splitline[0].length > 0.0 
            newline << splitline[1] if splitline[1].length > 0.0 
          else 
            newline << line
          end
        end
        divline = newline
      end
      if (hit3 == true) then
        newline = []
        divline.each do |line|
          if cross?(line, line3) then
            splitline = cutting(line, line3)
            newline << splitline[0] if splitline[0].length > 0.0 
            newline << splitline[1] if splitline[1].length > 0.0 
          else 
            newline << line
          end
        end
        divline = newline
      end
    end

    return divline
  end

end

